from __future__ import absolute_import, division, print_function, unicode_literals

import logging

from boto3.dynamodb.conditions import Attr, Key as ConditionKey
from boto3.dynamodb.types import Binary

logger = logging.getLogger(__name__)


class ItemNotFound(Exception):
    pass


class DynamoDbStorageService(object):

    @property
    def binary_type(self):
        # DynamoDB wont put `str` => `Binary` in python2
        # So we need to advertise how we'd like the encryption
        #  service to wrap it's binary data.
        return Binary

    def __init__(self, session, table_name):
        self.dynamodb = session.resource('dynamodb')
        self.table_name = table_name
        self._secrets_table = None

    @property
    def secrets_table(self):
        if not self._secrets_table:
            self._secrets_table = self.dynamodb.Table(self.table_name)
        return self._secrets_table

    def list_one(self, name):
        response = self.secrets_table.scan(
            FilterExpression=Attr("name").eq(name),
            ProjectionExpression="#N, version",
            ExpressionAttributeNames={"#N": "name"}
        )
        return [
            self._unwrap_doc(item)
            for item in response['Items']
        ]

    def list_all(self):
        response = self.secrets_table.scan(
            ProjectionExpression="#N, version",
            ExpressionAttributeNames={"#N": "name"}
        )
        return [
            self._unwrap_doc(item)
            for item in response['Items']
        ]

    def delete_one(self, name, version):
        self.secrets_table.delete_item(Key={
            'name': name,
            'version': self._wrap_int(version)
        })

    def put_one(self, name, version, ciphertext):
        data = {
            'name': name,
            'version': self._wrap_int(version),
        }
        data.update(ciphertext)
        self.secrets_table.put_item(
            Item=data,
            ConditionExpression=Attr('name').not_exists()
        )

    def get_one(self, name, version=None):
        if not version:
            return self._get_latest_secret(name)
        else:
            return self._get_versioned_secret(name, version)

    def _get_latest_secret(self, secret_name):
        # do a consistent fetch of the credential with the highest version
        response = self.secrets_table.query(
            Limit=1,
            ScanIndexForward=False,
            ConsistentRead=True,
            KeyConditionExpression=ConditionKey("name").eq(secret_name)
        )
        if response["Count"] == 0:
            return
        return self._unwrap_doc(response["Items"][0])

    def _get_versioned_secret(self, secret_name, version):
        version = self._wrap_int(version)
        response = self.secrets_table.get_item(Key={"name": secret_name, "version": version})
        if "Item" not in response:
            return
        return self._unwrap_doc(response["Item"])

    @classmethod
    def _unwrap_doc(cls, obj):
        new_obj = {
            k: (v.value if isinstance(v, Binary) else v)
            for k, v in obj.items()
        }
        new_obj['version'] = cls._unwrap_int(new_obj['version'])
        return new_obj

    INT_FMT = '019d'

    @classmethod
    def _wrap_int(cls, i):
        if not isinstance(i, int):
            raise TypeError('Expected int, got %s instead' % type(i))
        return format(i, cls.INT_FMT)

    @classmethod
    def _unwrap_int(cls, txt):
        try:
            return int(txt, 10)
        except ValueError:
            raise RuntimeError('Could not unwrap as integer: %s' % txt)

    def setup(self, read_capacity=1, write_capacity=1):
        table_names = {t.name for t in self.dynamodb.tables.all()}
        if self.table_name in table_names:
            raise SetupError("Credential Store table already exists")

        logger.info('creating table "%s"...', self.table_name)
        self.dynamodb.create_table(
            TableName=self.table_name,
            KeySchema=[
                {
                    "AttributeName": "name",
                    "KeyType": "HASH",
                },
                {
                    "AttributeName": "version",
                    "KeyType": "RANGE",
                }
            ],
            AttributeDefinitions=[
                {
                    "AttributeName": "name",
                    "AttributeType": "S",
                },
                {
                    "AttributeName": "version",
                    "AttributeType": "S",
                },
            ],
            ProvisionedThroughput={
                "ReadCapacityUnits": read_capacity,
                "WriteCapacityUnits": write_capacity,
            }
        )
        logger.debug('Waiting for table to be created...')
        self.dynamodb.get_waiter('table_exists').wait(TableName=self.table_name)

    def __repr__(self):
        return 'DynamoDbStorageService(table_name=%s)' % self.table_name


class SetupError(RuntimeError):
    pass



from __future__ import absolute_import, division, print_function

import codecs
import grp
import json
import logging
import os
import pwd

import click
import jinja2.sandbox

import credsmash.api
from .util import read_many, shell_quote, parse_manifest, detect_format
from .cli import main

logger = logging.getLogger(__name__)


class CachingProxy(object):
    def __init__(self, getter, key_fmt):
        self._getter = getter
        self._key_fmt = key_fmt
        self._data = {}

    def __getitem__(self, key):
        if key not in self._data:
            if isinstance(key, tuple):
                lookup_key = self._key_fmt.format(*key)
            else:
                lookup_key = self._key_fmt.format(key)
            logger.debug('key=%s lookup_key=%s', key, lookup_key)
            res = self._getter(lookup_key)
            self._data[key] = res
            return res
        return self._data[key]


@main.command('render-template')
@click.argument('template', type=click.File(mode='r', encoding='utf-8'))
@click.argument('destination', type=click.File(mode='w', encoding='utf-8'))
@click.option('--obj-name', default='secrets',
              help='The variable/object name provided to the template')
@click.option('--key-fmt', default='{0}',
              help='Re-use templates by tweaking which variable it maps to- '
                   'eg, "dev.{0}" converts {{secrets.potato}} to the secret "dev.potato"')
@click.option('--data-file', type=click.File(mode='rb'),
              help="Source from data file instead of credential store "
                   "(useful for caching/testing)")
@click.option('--data-file-format', default=None)
@click.pass_context
def cmd_render_template(
        ctx, template, destination,
        obj_name='secrets', key_fmt='{0}',
        data_file=None, data_file_format=None
):
    """
    Render a configuration template....
    """
    if data_file:
        if not data_file_format:
            data_file_format = detect_format(data_file, 'json')
        data = read_many(data_file, data_file_format)
        secrets = CachingProxy(lambda key: data[key], key_fmt)
    else:
        secrets = CachingProxy(lambda key: credsmash.api.get_secret(
            ctx.obj.storage_service,
            ctx.obj.key_service,
            key,
        ), key_fmt)

    env = _make_env()
    output = env.from_string(template.read()).render(**{
        obj_name: secrets
    })
    destination.write(output)


@main.command('render-templates')
@click.argument('manifest', type=click.File(mode='r', encoding='utf-8'))
@click.option('--manifest-format', default=None)
@click.option('--obj-name', default='secrets',
              help='The variable/object name provided to the template')
@click.option('--key-fmt', default='{0}',
              help='Re-use templates by tweaking which variable it maps to- '
                   'eg, "dev.{0}" converts {{secrets.potato}} to the secret "dev.potato"')
@click.option('--data-file', type=click.File(mode='rb'),
              help="Source from data file instead of credential store "
                   "(useful for caching/testing)")
@click.option('--data-file-format', default=None)
@click.pass_context
def cmd_render_template(
        ctx, manifest, manifest_format=None,
        obj_name='secrets', key_fmt='{0}',
        data_file=None, data_file_format=None
):
    """
    Render multiple configuration templates - reads from a manifest file.
    """
    if data_file:
        if not data_file_format:
            data_file_format = detect_format(data_file, 'json')
        data = read_many(data_file, data_file_format)
        secrets = CachingProxy(lambda key: data[key], key_fmt)
    else:
        secrets = CachingProxy(lambda key: credsmash.api.get_secret(
            ctx.obj.storage_service,
            ctx.obj.key_service,
            key,
        ), key_fmt)

    env = _make_env()
    if not manifest_format:
        manifest_format = detect_format(manifest, 'json')
    for entry in parse_manifest(manifest, manifest_format):
        with codecs.open(entry['source'], 'r', encoding='utf-8') as template, \
                codecs.open(entry['destination'], 'w', encoding='utf-8') as destination:
            output = env.from_string(template.read()).render(**{
                obj_name: secrets
            })
            destination.write(output)

        if 'mode' in entry:
            os.chmod(
                entry['destination'],
                entry['mode']
            )

        if 'owner' in entry and 'group' in entry:
            os.chown(
                entry['destination'],
                pwd.getpwnam(entry['owner']).pw_uid,
                grp.getgrnam(entry['group']).gr_gid
            )
        logger.info('Rendered template="%s" destination="%s"', entry['source'], entry['destination'])


def _make_env():
    env = jinja2.sandbox.SandboxedEnvironment(
        trim_blocks=True,
        lstrip_blocks=True,
        autoescape=False,
    )
    env.filters['sh'] = shell_quote
    env.filters['jsonify'] = json.dumps
    return env



from __future__ import absolute_import, division, print_function, unicode_literals

from credsmash.crypto import open_secret
from credsmash.util import ItemNotFound


def get_secret(storage_service, key_service, secret_name, version=None):
    ciphertext = storage_service.get_one(secret_name, version=version)
    if not ciphertext:
        raise ItemNotFound("Item {'name': '%s', 'version': %s} couldn't be found." % (secret_name, version))
    return open_secret(key_service, ciphertext)
